//
// Created by 陈展 on 24-5-30.
//

#include <utility>

#include "executor_abstract.h"
#ifndef RMDB_EXECUTION_AGG_H
#define RMDB_EXECUTION_AGG_H

#endif  // RMDB_EXECUTION_AGG_H

struct VectorValueHasher {
  std::size_t operator()(const std::vector<Value>& vec) const {
    std::size_t seed = vec.size();
    for(const auto& val : vec) {
      seed ^= val.hash();
    }
    return seed;
  }
};


class AggExecutor : public AbstractExecutor {
 private:
  std::unique_ptr<AbstractExecutor> prev;  // 上游执行器
  std::vector<AggColMeta> aggMetas;                    // 聚合函数
  std::vector<AggColMeta> groupByCols;                   // 分组字段
  std::vector<HavingConditon> havingConds;                // having条件
  std::unordered_map<std::vector<Value>,std::vector<AggValue>,VectorValueHasher> havingRes; //每个分组的having结果
  std::vector<ColMeta> colMetas;            //cols
  std::map<std::string,char *> dataMap;             // 字段名和数据的映射
  size_t tupleLength;                               // 每个元组的长度
  //map key左边是group by的列的值，右边是聚合函数运算前的值
  std::unordered_map<std::vector<Value>, std::vector<AggValue>,VectorValueHasher> groupByResults; // 分组后的聚合结果
  std::vector<std::vector<Value>> resultKeys;     // 聚合结果的键列表
  size_t resultIndex;                             // 当前结果索引
  Context *context;                                 // 上下文信息
  bool count0 = false;
  SmManager* sm_manager_;

 public:
  AggExecutor(std::unique_ptr<AbstractExecutor> prev,
              std::vector<AggCol> aggCols,
              std::vector<TabCol> gCols,
              std::vector<HavingConditon> hConds,
              SmManager* sm_manager_,
              Context *context)
      : prev(std::move(prev)), havingConds(std::move(hConds)), context(context),sm_manager_(sm_manager_) {
      TabMeta tab;
      dataMap.clear();
      if(!aggCols.empty()) {
        tab = sm_manager_->db_.get_table(aggCols[0].col.tab_name);
      }else {
        tab = sm_manager_->db_.get_table(gCols[0].tab_name);
      }
      int offset = 0;
      //下面的offset是返回给上层的offset
    //处理group by字段
      for(auto &gCol : gCols){
        AggColMeta aggMeta;//给聚合用的
        ColMeta colMeta;//给投影用的
        //取出表中字段原来的元信息
        colMeta = *tab.get_col(gCol.col_name);
        aggMeta.col = colMeta;//获取原数据
        //给偏移量赋值
        colMeta.offset = offset;
        colMeta.tab_name = gCol.tab_name;
        colMeta.name = gCol.col_name;
        colMeta.ailas = gCol.alias;
        colMetas.push_back(colMeta);
        groupByCols.push_back(aggMeta);
        offset+=colMeta.len;
      }
    //处理聚集函数
      for(auto &aCol : aggCols){
        AggColMeta aggMeta;
        ColMeta colMeta;
        //将TabCol转化为ColMeta
        //取出表中字段原来的元信息
        aggMeta.ag_type = aCol.AggType;
        //将aggMeta提取到Colmeta
        colMeta.offset = offset;
        if(aCol.col.col_name != "*"){
          aggMeta.col = *tab.get_col(aCol.col.col_name);
          //如果是string且是非Count类型，报错
          if(aggMeta.col.type == TYPE_STRING && aggMeta.ag_type == ast::AggFuncType::SUM ){
                throw InternalError("String type can only be used in COUNT function");
          }
          colMeta.tab_name = aCol.col.tab_name;
          colMeta.name = aCol.col.col_name;
          colMeta.ailas = aCol.col.alias;
        }else {
          aggMeta.ag_type = ast::AggFuncType::COUNT_ALL;
          colMeta.tab_name = "*";
          colMeta.name = "*";
          colMeta.ailas = aCol.col.alias;
        }
        GetCol(aggMeta, colMeta);
        colMetas.push_back(colMeta);
        offset+=colMeta.len;
        aggMetas.push_back(aggMeta);
      }
      //处理havingConds
      for(auto &cond : havingConds) {
        AggColMeta aggMeta;
        if(cond.is_agg) {
          aggMeta.ag_type = cond.ag_type;
          if(cond.lhs_col.col_name != "*") {
            aggMeta.col = *tab.get_col(cond.lhs_col.col_name);
            //如果是string且是非Count类型，报错
            if(aggMeta.col.type == TYPE_STRING && aggMeta.ag_type != ast::AggFuncType::COUNT && aggMeta.ag_type != ast::AggFuncType::COUNT_ALL){
              throw InternalError("String type can only be used in COUNT function");
            }
          }else {
            aggMeta.ag_type = ast::AggFuncType::COUNT_ALL;
          }
        }else {

        }
        cond.agg_col = aggMeta;
      }
      tupleLength = GetTupleLen(aggMetas,groupByCols);
      resultIndex = 0;
    }

    bool is_end() const override {
      return resultIndex >= resultKeys.size();
    }


    const std::vector<ColMeta> &cols() const override {
      return this->colMetas;
    }

    void beginTuple() override {
      groupByResults.clear();
      dataMap.clear();
      resultKeys.clear();
      havingRes.clear();
      count0 = false;
      resultIndex = 0;
      prev->beginTuple();
      //如果没有group by
      if(groupByCols.empty()){
        std::vector<Value> key;
        groupByResults[key] = std::vector<AggValue>(aggMetas.size());
        resultKeys.push_back(key);
      }

      if(prev->is_end()){
        //如果是count0的情况
        bool all_cnt = true;
        for(auto agg : aggMetas){
          if(agg.ag_type != ast::AggFuncType::COUNT && agg.ag_type != ast::AggFuncType::COUNT_ALL){
            all_cnt = false;
            break;
          }
        }
        if(all_cnt){
          count0 = true;
        }else{
          resultIndex++;
        }
        return;
      }
      //一次性取出全部数据
      while (!prev->is_end()) {
        auto tuple = prev->Next();
        ProcessTuple(tuple);
        dataMap.clear();
        prev->nextTuple();
      }
      //找到第一个满足条件的turple
       while (resultIndex < resultKeys.size() && !havingConds.empty() && !Check_Having_Conditions(resultKeys[resultIndex])){
          ++resultIndex;
       }
    }

    void nextTuple() override {
    do {
      resultIndex++;
    }while(resultIndex < resultKeys.size() && !havingConds.empty() && !Check_Having_Conditions(resultKeys[resultIndex]));
    }

    std::unique_ptr<RmRecord> Next() override {
      if (is_end()) {
        return nullptr;
      }
      if(count0){
        auto res = std::make_unique<RmRecord>(sizeof(int));
        *(int *)res->data = 0;
        return res;
      }
      auto &key = resultKeys[resultIndex];
      auto &result = groupByResults[key];
      auto res = std::make_unique<RmRecord>(tupleLength);
      // 将键和聚合结果合并到 res
      MergeKeyAndResult(key, result, res);
      return res;
    }

    size_t tupleLen() const override {
      return tupleLength;
    }

    Rid &rid() override {
      return prev->rid();
    }

   private:


    void GetCol(AggColMeta &aCol,ColMeta &col) const {
      switch(aCol.ag_type){
        case ast::AggFuncType::SUM:
          col.type = aCol.col.type;
          col.len = aCol.col.len;
          break;
        case ast::AggFuncType::COUNT:
          col.type = TYPE_INT;
          col.len = sizeof(int);
          break;
        case ast::AggFuncType::COUNT_ALL:
          col.type = TYPE_INT;
          col.len = sizeof(int);
          break;
        case ast::AggFuncType::AVG:
          col.type = TYPE_FLOAT;
          col.len = sizeof(float);
          break;
        case ast::AggFuncType::MIN:
          col.type = aCol.col.type;
          col.len = aCol.col.len;
          break;
        case ast::AggFuncType::MAX:
          col.type = aCol.col.type;
          col.len = aCol.col.len;
          break;
      }
    }


    size_t GetTupleLen(std::vector<AggColMeta> &aggMetas, std::vector<AggColMeta> &groupByCols) {
      size_t len = 0;
      for (auto &col : groupByCols) {
        len += col.col.len;
      }
      for (auto &aCol : aggMetas) {
        switch (aCol.ag_type) {
          case ast::AggFuncType::SUM:
            len += aCol.col.len;
            break;
          case ast::AggFuncType::COUNT:
            len += sizeof(int);
            break;
          case ast::AggFuncType::COUNT_ALL:
            len += sizeof(int);
            break;
          case ast::AggFuncType::AVG:
            len += sizeof(float);
            break;
          case ast::AggFuncType::MIN:
            len += aCol.col.len;
            break;
          case ast::AggFuncType::MAX:
            len += aCol.col.len;
            break;
        }
      }
      return len;
    }

    void ProcessTuple(std::unique_ptr<RmRecord> &tuple) {
      //读取record数据到map
      //组装group by的key
      std::vector<Value> key;
      //获取数据
      for(auto gCol : groupByCols){
        dataMap[gCol.col.name] = tuple->data + gCol.col.offset;
        //将一个tuple里的每条记录的每个字段数据取出来依次装到key里面
        //tuple的偏移量要重新拿，不是gCol的偏移量
        key.emplace_back(gCol.col.type,tuple->data + gCol.col.offset,gCol.col.len);
      }
      for(auto aCol : aggMetas){
        dataMap[aCol.col.name] = tuple->data + aCol.col.offset;
      }
      for(auto & havingCond : havingConds) {
        dataMap[havingCond.agg_col.col.name] = tuple->data + havingCond.agg_col.col.offset;
      }
      if (groupByResults.find(key) == groupByResults.end()) {//没找到聚合组
        groupByResults[key] = std::vector<AggValue>(aggMetas.size());
        resultKeys.push_back(key);
      }
      //初始化having
      if(!havingConds.empty()) {
        if(havingRes.find(key) == havingRes.end()) {
          havingRes[key] = std::vector<AggValue>(havingConds.size());
        }
      }

      std::vector<AggValue>& aggResult = groupByResults[key];
      // 对于每一个 aggMetas，计算其值，并将其添加到 aggResult 中
      for (size_t i = 0 ;i<aggMetas.size();i++) {
        // 计算 aggMeta 的值
        CalculateAggValue(aggMetas[i],aggResult[i]);
      }
      //对于having里的每一个聚集函数计算其值
      std::vector<AggValue>& havingResult = havingRes[key];
      for(size_t i=0;i<havingConds.size();i++) {
        auto &cond = havingConds[i];
        CalculateAggValue(cond.agg_col,havingResult[i]);
      }
    }

    void CalculateAggValue(const AggColMeta &aggMeta,AggValue& res) {
      // 从 dataMap 中取出数据
      char *data = dataMap[aggMeta.col.name];
      // 计算 aggMeta 的值
      if (aggMeta.ag_type == ast::AggFuncType::COUNT || aggMeta.ag_type == ast::AggFuncType::COUNT_ALL) {
        res.type = TYPE_INT;
        //题目要求空字符串也计数，count *和count(字段) 应该没有区别
        // if (aggMeta.ag_type == ast::AggFuncType::COUNT_ALL && aggMeta.col.type == TYPE_STRING){
        //   if(*(char *)data != '\0'){
        //     res.int_val++;
        //   }
        // }else{
          res.int_val++;
        // }
      } else if (aggMeta.ag_type == ast::AggFuncType::SUM) {
        if (aggMeta.col.type == TYPE_INT) {
          res.type = TYPE_INT;
          res.int_val += *(int *)data;
        } else if (aggMeta.col.type == TYPE_FLOAT) {
          res.type = TYPE_FLOAT;
          res.float_val += *(float *)data;
        }
      }else if(aggMeta.ag_type == ast::AggFuncType::MAX){
        if(aggMeta.col.type == TYPE_INT){
          if(!res.maxFlag){
            res.int_val = *(int *)data;
            res.maxFlag = true;
          }
          res.type = TYPE_INT;
          res.int_val = std::max(res.int_val,*(int *)data);
        }else if(aggMeta.col.type == TYPE_FLOAT){
          if(!res.maxFlag){
                  res.float_val = *(float *)data;
                  res.maxFlag = true;
          }
          res.type = TYPE_FLOAT;
          res.float_val = std::max(res.float_val,*(float *)data);
        }else if (aggMeta.col.type == TYPE_STRING) {
          if(!res.maxFlag) {
            res.str_val = std::string(data,aggMeta.col.len);
                res.maxFlag = true;
          }
                res.type = TYPE_STRING;
          res.str_val = std::max(res.str_val,std::string(data,aggMeta.col.len));
        }
    }else if(aggMeta.ag_type == ast::AggFuncType::MIN){
        if(aggMeta.col.type == TYPE_INT){
          if(!res.minFlag){
            res.int_val = *(int *)data;
            res.minFlag = true;
          }
          res.type = TYPE_INT;
          res.int_val = std::min(res.int_val,*(int *)data);
        }else if(aggMeta.col.type == TYPE_FLOAT){
          if(!res.minFlag){
          res.float_val = *(float *)data;
          res.minFlag = true;
          }
          res.type = TYPE_FLOAT;
          res.float_val = std::min(res.float_val,*(float *)data);
        }else if(aggMeta.col.type == TYPE_STRING) {
          if(!res.minFlag) {
            res.str_val = std::string(data,aggMeta.col.len);
            res.minFlag = true;
          }
          res.type = TYPE_STRING;
          res.str_val = std::min(res.str_val,std::string(data,aggMeta.col.len));
        }
    }else{
      throw std::runtime_error("Unknown AggFuncType");
    }
 }


    void MergeKeyAndResult(std::vector<Value> &key, std::vector<AggValue> &result, std::unique_ptr<RmRecord> &res) {

      size_t offset = 0;
      for (auto &val : key) {
        size_t len = val.get_len();
        memcpy(res->data + offset, val.get_data(),len);
        offset += len;
      }
      for (auto &val : result) {
        size_t len = val.get_len();
        memcpy(res->data + offset, val.get_data(),len);
        offset += len;
      }
    }
    bool Check_Having_Conditions(const std::vector<Value>& key) {
        //对每个cond验证
        for(int i=0;i<havingConds.size();i++) {
          auto hRes = havingRes[key];
          //hres里有值，cond里有条件，直接验证，只考虑右边为常量
          ColType left_t = hRes[i].type;
          char *left = hRes[i].get_data();
          ColType right_t = havingConds[i].rhs_val.type;
          char *right = havingConds[i].rhs_val.get_data();
          int cmp = 0;
          if(left_t == TYPE_INT && right_t == TYPE_FLOAT) {
            char* left_float = new char[4];
            *(float*)left_float = *(int*)left;
            cmp = ix_compare(left_float, right, right_t, hRes[i].get_len());
            delete[] left_float;
          }else if(left_t == TYPE_FLOAT && right_t == TYPE_INT) {
            char* right_float = new char[4];
            *(float*)right_float = *(int*)right;
            cmp = ix_compare(left, right_float, left_t, hRes[i].get_len());
            delete[] right_float;
          }else {
            cmp = ix_compare(left, right, right_t, hRes[i].get_len());
          }
          if (havingConds[i].op == OP_EQ) {
            if(cmp != 0) return false;
          } else if (havingConds[i].op == OP_NE) {
            if(cmp == 0) return false;
          } else if (havingConds[i].op == OP_LT) {
            if(cmp >= 0) return false;
          } else if (havingConds[i].op == OP_GT) {
            if(cmp <= 0) return false;
          } else if (havingConds[i].op == OP_LE) {
            if(cmp > 0) return false;
          } else if (havingConds[i].op == OP_GE) {
            if(cmp < 0) return false;
          } else {
            throw std::runtime_error("Unexpected op type");
          }
        }
      return true;
      }
  };


