/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class ProjectionExecutor : public AbstractExecutor {
   private:
    std::unique_ptr<AbstractExecutor> prev_;        // 投影节点的儿子节点
    std::vector<ColMeta> cols_;                     // 需要投影的字段
    size_t len_;                                    // 字段总长度
    std::vector<size_t> sel_idxs_; //表示要投影字段在cols的下标

   public:
    ProjectionExecutor(std::unique_ptr<AbstractExecutor> prev, const std::vector<TabCol> &sel_cols,bool isAgg) {
        prev_ = std::move(prev);
        size_t curr_offset = 0;
        auto prev_cols = prev_->cols();
        // for(auto &pc : prev_cols) {
        //     if(pc.ailas.empty()) {
        //         pc.ailas = pc.name;
        //     }
        // }
        if(!isAgg) {
          for (auto &sel_col : sel_cols) {
            auto pos = get_col(prev_cols, sel_col);
            sel_idxs_.push_back(pos - prev_cols.begin());
            auto col = *pos;
            col.offset = curr_offset;
            curr_offset += col.len;
            cols_.push_back(col);
          }
        }else {
            for (auto &sel_col : sel_cols) {
                auto pos = get_agg_col(prev_cols, sel_col);
                sel_idxs_.push_back(pos - prev_cols.begin());
                auto col = *pos;
                col.offset = curr_offset;
                curr_offset += col.len;
                cols_.push_back(col);
            }
        }
        len_ = curr_offset;

    }

    const std::vector<ColMeta> &cols() const override{
        return this->cols_;
    };

    size_t tupleLen() const override {
      return len_;
    }

    bool is_end() const override{
       return prev_->is_end();
    }

    void beginTuple() override {
        prev_->beginTuple();
    }

    void nextTuple() override {
        prev_->nextTuple();
    }

    std::unique_ptr<RmRecord> Next() override {
//        assert(is_end());
        auto preCols = prev_->cols(); //上一个算子传来的数据的字段
        auto preRec = prev_->Next(); //上一个算子传来的记录
        std::unique_ptr<RmRecord> record = std::make_unique<RmRecord>(len_);
        for(size_t i=0; i < sel_idxs_.size(); i++){
            auto &selCol = cols_[i]; //cols与sel_idx的长度相同
            auto &preCol = preCols[sel_idxs_[i]];
            memcpy(record->data + selCol.offset, preRec->data + preCol.offset, selCol.len);
        }
        return record;

    }

    Rid &rid() override { return _abstract_rid; }
};
